# Overview
`Projector` is a logic program rewriting tool designed to split a single rule into multiple rules with the goal that the number of variables present in each new rule is smaller than that of the original.

`Projector-Predictor` mode is also available and can be activiated with the --predictor flag.

# Installation
The only dependency for `Projector` is `clingo` 5.3.0 (other version may work, but we have tested for this version). It must be installed with `Python` 2.7 enabled and the binary must be accessible to your python path.

The easiest way to install clingo and all other dependencies into your python path is through `Anaconda`:
1. Install `Anaconda` for `Python` 2.7. `Anaconda` can be downloaded [here](https://www.anaconda.com/download/). Follow the instructions provided on the site for installation.
2. Install `clingo` 5.3.0 by running `conda install -c potassco clingo=5.3.0`. Release information for clingo can be found [here](https://github.com/potassco/clingo/releases).
3. Install `networkx` 2.2 by running `conda install networkx=2.2`.
4. Install `numpy` 1.14.2 by running `conda install numpy=1.14.2`.

# Running Projector
To run projector, simply execute `python projector.py (encoding-file)`.

Detailed usage info can be found by running `python projector.py --help`.

## Flags
There are many flags intended for debugging purposes. The following covers the flags relevant to most users.

`--instance` Specifies an additional file to include in grounding/solving. Projector will not rewrite these files. This flag may be provided multiple times if multiple instance files are to be used.

`--output` The output mode of the program (ground|solve|rewrite). Option `ground` will provide statistical info for the grounding process only. Option `solve` will provide statistical info for both grounding and solving. Option `rewrite` (default) will output the rewritten program.

`--predictor` Enables the use of [Predictor](https://www.unomaha.edu/college-of-information-science-and-technology/natural-language-processing-and-knowledge-representation-lab/software/predictor.php) for variable heuristics (deciding whether or not to perform a projection)

`--key` specifies the predicate keys to use with predictor (predictor must be enabled; format: p/3[1,3])

`--version` Displays the current version of `Projector`.

`--help` Displays help info. See below for usage

## Full Usage (--help)
```
usage: projector.py [-h] [--instance INSTANCE] [--output OUTPUT]
                    [--debug DEBUG] [--max_atoms MAX_ATOMS]
                    [--restrict RESTRICT] [--no_normalize] [--limit LIMIT]
                    [--debug_after DEBUG_AFTER] [--random RANDOM]
                    [--random_rules RANDOM_RULES] [--oracle] [--no_rewrite]
                    [--heuristic HEURISTIC] [--aggregates] [--old_mode]
                    [--conservative] [--predictor] [--key KEY] [--version]
                    encoding [encoding ...]

Rewrites ASP logic program

positional arguments:
  encoding              gringo source files

optional arguments:
  -h, --help            show this help message and exit
  --instance INSTANCE   adds an additional file to include in
                        grounding/solving (will not be rewritten)
  --output OUTPUT       the output mode of the program [ground|solve|rewrite]
  --debug DEBUG         enables pausing after projection & debug trace
                        messages, value controls the level, 0 disables
  --max_atoms MAX_ATOMS
                        the maximum number of atoms a variable can be shared
                        with in a rule while still performing projection, -1
                        for no max atoms
  --restrict RESTRICT   the restrict mode of the program
  --no_normalize        normalizes projected rules (NOTE: breaks caching for
                        differently named variables)
  --limit LIMIT         limits the number of projections performed, stopping
                        projection after encountering the number specified
  --debug_after DEBUG_AFTER
                        enables debug mode after specified number of
                        projections
  --random RANDOM       provides a seed for randomness on equal heuristics, 0
                        to disable and always choose the first encountered
                        (default)
  --random_rules RANDOM_RULES
                        provides a seed for random choice selection on whether
                        or not to perform a rewrite, 0 to disable (default)
  --oracle              predicts projections based on grounding size (will
                        only improve solve time; grounding will occur many
                        times)
  --no_rewrite          disables all rewriting and just parses the given
                        program
  --heuristic HEURISTIC
                        the variable selection heuristic to use
                        [variable|ascending|descending]
  --aggregates          enable projection on aggregates
  --old_mode            sets certain changes back to the state they were in
                        during the paper (currently just beta literal removal
                        from replacement rule)
  --conservative        enables conservative projection mode
  --predictor           enables the use of Predictor for variable heuristics
  --key KEY             specifies the predicate keys to use with predictor
                        (predictor must be enabled; format: p/3[1,3])
  --version             prints the version of Projector
```

## Examples
`python projector.py examples/male_cousin.lp`
`python projector.py examples/male_cousin.lp --predictor`

# The MIT License (MIT)

Copyright 2018 Nicholas Hippen

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
of the Software, and to permit persons to whom the Software is furnished to do
so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.