# DEPRECATED

from clingo_ast_util import is_predicate_in_positive_body, is_predicate_in_negative_body

positive_label = '+'
negative_label = '-'

class DependencyGraph():

    def __init__(self):
        self.edge_dict = {}
    
    def add_edge(self, from_p ,to_p):
        if from_p not in self.edge_dict:
            self.edge_dict[from_p] = set()
        self.get_edges(from_p).add(to_p)
    
    def get_edges(self, from_p):
        return self.edge_dict.get(from_p)

    def nodes(self):
        return self.edge_dict.keys()
    
    def fill_order(self, from_p, visited, stack):
        visited.add(from_p)
        if from_p in self.edge_dict:
            for to_p in self.get_edges(from_p):
                if to_p not in visited:
                    self.fill_order(to_p, visited, stack)
        stack = stack.append(from_p)

    def get_transpose(self):
        transpose = DependencyGraph()

        for from_p in self.nodes():
            for to_p in self.get_edges(from_p):
                transpose.add_edge(to_p, from_p)
        return transpose
    
    def dfs_util(self, node, visited, component):
        visited.add(node)
        component.add(node)
        if node in self.edge_dict:
            for to_p in self.get_edges(node):
                if to_p not in visited:
                    self.dfs_util(to_p, visited, component)

    def strongly_connected_components(self):
        stack = []
        visited = set()
        for from_p in self.nodes():
            if from_p not in visited:
                self.fill_order(from_p, visited, stack)
        
        transpose = self.get_transpose()
        visited = set()
        scc = []
        while stack:
            node = stack.pop()
            if node not in visited:
                component = set()
                transpose.dfs_util(node, visited, component)
                scc.append(tuple(component)) # Components should be immutable
        return scc
    
    def clear(self):
        self.edge_dict = {}
    
    def __str__(self):
        graph_str = ''
        for from_p in self.nodes():
            graph_str += '%s ---> %s\n' % (from_p, self.get_edges(from_p))
        return graph_str

class ComponentGraph(DependencyGraph):

    def __init__(self):
        self.edge_labels = {}
        return super().__init__()

    def add_edge(self, from_p, to_p, label):
        self.edge_labels[from_p] = label
        return super().add_edge(from_p, to_p)

    def get_label(self, from_p):
        return self.edge_labels.get(from_p)

    def clear(self):
        self.edge_labels = {}
        return super().clear()

    def update(self, data_holder):
        self.clear()
        scc = data_holder.dg.strongly_connected_components()
        for component_A in scc:
            module = data_holder.get_module(component_A)
            print('%s: %s' % (component_A, module))
            for component_B in scc:
                if component_A == component_B:
                    continue
                negative = False
                positive = False
                for rule in module:
                    # We know that a predicate from component_A is in the head already.
                    # Now we just check if a predicate from component_B is in the positive body or negative body.
                    if any(is_predicate_in_positive_body(rule, pred_B) for pred_B in component_B):
                        self.add_edge(component_B, component_A, positive_label)
                        positive = True
                    elif any(is_predicate_in_negative_body(rule, pred_B) for pred_B in component_B):
                        negative = True
                if negative and not positive:
                    self.add_edge(component_B, component_A, negative_label)

    def __str__(self):
        graph_str = ''
        for from_p in self.nodes():
            graph_str += '%s -[%s]-> %s\n' % (from_p, self.get_label(from_p), self.get_edges(from_p))
        return graph_str
