import random, time, json, sys
import clingo
import pr_settings
from pr_stdio import dprint, print_settings
from pr_transformer import Transformer
# sys.path.append('../predictor/') # TODO Not this
import clingo_ast_util
from predictor import Predictor
from pd_utils import KeyList, parse_key
import clingo_ast_util

class App:
    def __init__(self):
        pass

    def create_control(self):
        control = clingo.Control(['--warn=none'])
        control.use_enumeration_assumption = False
        if pr_settings.args.instance:
            for instance_path in pr_settings.args.instance:
                control.load(instance_path)
        return control

    def run(self):
        dprint('~~~~[DEBUG ENABLED]~~~~')
        print_settings()
        smallest_grounded_rules = float('inf') # For oracle
        smallest_control = float('inf') # For oracle
        parse = True
        loop_count = 0
        while parse:
            loop_count += 1
            self.control = self.create_control()
            if not pr_settings.args.oracle:
                parse = False
            with self.control.builder() as b:
                parse_start = time.time()
                if pr_settings.args.predictor: # TODO We should be able to use the same clingo instance to avoid parsing twice
                    key_list = KeyList()
                    if pr_settings.args.key:
                        for key_str in pr_settings.args.key:
                            key_list.add_key(parse_key(key_str))
                    predictor = Predictor(key_list)
                    for encoding_path in pr_settings.args.encoding:
                        predictor.load_program(encoding_path)
                    if pr_settings.args.instance:
                        for instance_path in pr_settings.args.instance:
                            predictor.load_program(instance_path)
                else:
                    predictor = None
                t = Transformer(b, predictor)
                clingo.parse_program(
                    read_encodings(),
                    lambda stm: t.prepare_stm(stm))
                t.build_program()
                parse_end = time.time()
                parse_time = parse_end - parse_start
            if pr_settings.args.oracle:
                ground_start = time.time()
                self.control.ground([('base', [])])
                ground_end = time.time()
                ground_time = ground_end - ground_start
                grounded_rules = self.control.statistics['problem']['lpStep']['rules']
                dprint('Grounding size WAS: %s' % smallest_grounded_rules)
                if loop_count == 1:
                    smallest_grounded_rules = grounded_rules
                    smallest_control = self.control
                    dprint('Base grounding size is %s' % smallest_grounded_rules)
                else:
                    dprint('Checking grounding size of %s' % grounded_rules)
                    if grounded_rules < smallest_grounded_rules:
                        smallest_grounded_rules = grounded_rules
                        smallest_control = self.control
                        pr_settings.answer_queue.append('y')
                        dprint('Changed grounding size to: %s' % smallest_grounded_rules)
                    else:
                        pr_settings.answer_queue.append('n')
                if pr_settings.oracle_done: # We've tried every first-selection projection
                    dprint('Oracle done.')
                    parse = False
                    self.control = smallest_control
                    break
                dprint('Answer queue: %s' % pr_settings.answer_queue)
        if pr_settings.args.output == 'rewrite':
            for r in t.stms:
                print(r)
        else:
            if not pr_settings.args.oracle:
                ground_start = time.time()
                self.control.ground([('base', [])])
                ground_end = time.time()
                ground_time = ground_end - ground_start
                # print('BEFORE:')
                # print(json.dumps(self.control.statistics, sort_keys=True, indent=2, separators=(',', ': ')))
            if pr_settings.args.output == 'solve':
                solve_start = time.time()
                ret = self.control.solve()
                solve_end = time.time()
                solve_time = solve_end - solve_start
                self.control.statistics['summary']['times']['py-solve'] = solve_time
                self.control.statistics['summary']['times']['py-gs'] = ground_time + solve_time
                self.control.statistics['summary']['times']['py-total'] = parse_time + ground_time + solve_time
                self.control.statistics['summary']['satisfiable'] = ret.satisfiable
            elif pr_settings.args.output == 'ground':
                self.control.statistics['summary']['times']['py-total'] = parse_time + ground_time
            else:
                raise Exception('Unknown output setting')
            self.control.statistics['summary']['times']['py-parse'] = parse_time
            self.control.statistics['summary']['times']['py-ground'] = ground_time
            proj_stats = {
                'total': t.total_projection_count
            }
            for num, count in t.projection_counts.items():
                proj_stats[str(num)] = count
            self.control.statistics['summary']['rewrites'] = {
                'projection': proj_stats
            }
            if len(pr_settings.answer_queue) > 0:
                self.control.statistics['summary']['choices'] = pr_settings.answer_queue
            print(json.dumps(self.control.statistics, sort_keys=True, indent=2, separators=(',', ': ')))

def read_encodings():
    full_prog = ''
    for encoding in pr_settings.args.encoding:
        full_prog += open(encoding).read()
    return full_prog
