import clingo.ast
import clingo_ast_util
import pr_settings
import time
from clingo_ast_util import ASTWrapper
from pr_stdio import dprint

alpha_total = 0
dict_total = 0

def alpha(transformer, target, var, context_rule=None):
    # dprint('alpha(%s,%s,%s,%s)' % (transformer, target, var, context_rule))
    global alpha_total
    # TODO Replace alpha_dict usage with separate one that only builds for var
    # start = time.time()
    alpha_dict = AlphaDictionaryBuilder(base_transformer=transformer, ignore_head=True, ignore_aggregates=type(target) != list).get_alpha_dict(target)
    if type(var) in [list, set]:
        alpha_literals = set()
        for v in var:
            if type(v) is ASTWrapper:
                wrapped_var = v
            else:
                wrapped_var = ASTWrapper(v)
            alpha_literals = alpha_literals.union(alpha_dict.get(wrapped_var, set()))
    else:
        wrapped_var = ASTWrapper(var)
        alpha_literals = alpha_dict.get(wrapped_var, set())
    # end = time.time()
    # dprint('Alpha Time: %f' % (end - start))
    # alpha_total += end - start
    # dprint('Total Alpha: %f' % alpha_total, 0)
    return alpha_literals

def beta(transformer, target, var, context_rule=None):
    alpha_literals = alpha(transformer, target, var)
    alpha_vars = clingo_ast_util.find_variables(alpha_literals)
    if context_rule is None:
        candidate_literals = alpha(transformer, target, alpha_vars)
    else:
        candidate_literals = clingo_ast_util.find_ast_type(context_rule.body, clingo.ast.ASTType.Literal)
    new_beta_literals = set()
    for candidate in candidate_literals:
        if candidate in alpha_literals:
            continue
        candidate_vars = clingo_ast_util.find_variables(candidate.ast)
        if candidate_vars.issubset(alpha_vars):
            new_beta_literals.add(candidate)
    beta_literals = alpha_literals.union(new_beta_literals)
    return beta_literals

dict_builder_cache = {}

class AlphaDictionaryBuilder(clingo_ast_util.ASTVisitor):

    def __init__(self, base_transformer, variable_counter=None, tree_scope=None, parent_scope=None, ignore_head=False, ignore_aggregates=True):
        self.base_transformer = base_transformer
        if variable_counter is None:
            self.variable_counter = VariableCounter()
        else:
            self.variable_counter = variable_counter
        if tree_scope is None:
            self.tree_scope = {}
        else:
            self.tree_scope = tree_scope
        self.tree_scope['parent_scope'] = parent_scope
        self.ignore_head = ignore_head
        self.ignore_aggregates = ignore_aggregates
        self.loaded_rule = None
    
    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return False
        return self.base_transformer == other.base_transformer and self.tree_scope == other.tree_scope and self.tree_scope.get('parent_scope', None) == other.tree_scope.get('parent_scope', None) and self.ignore_head == other.ignore_head and self.ignore_aggregates == other.ignore_aggregates

    def __hash__(self):
        # TODO Don't do this probably...
        return hash('hash[base_transformer=%s, tree_scope=%s, parent_scope=%s, ignore_head=%s, ignore_aggregates=%s]' % (self.base_transformer, self.tree_scope, self.tree_scope.get('parent_scope', None), self.ignore_head, self.ignore_aggregates))

    def load_from_cache(self, rule):
        if self not in dict_builder_cache:
            dprint('No alpha dict in cache')
            return None
        cached_dict_builder = dict_builder_cache[self]
        if cached_dict_builder.loaded_rule != rule:
            dprint('Loaded alpha dict does not match rule')
            return None
        dprint('Alpha dict loaded from cache')
        return cached_dict_builder.variable_counter.variable_literal_dict

    def get_alpha_dict(self, rule):
        global dict_total, dict_builder_cache
        cached_alpha_dict = self.load_from_cache(rule)
        if cached_alpha_dict is not None:
            return cached_alpha_dict
        self.loaded_rule = rule
        # dprint('__init__(%s, variable_counter=%s, tree_scope=%s, parent_scope=%s, ignore_head=%s, ignore_aggregates=%s)' % (self.base_transformer, self.variable_counter, self.tree_scope, self.tree_scope.get('parent_scope', None), self.ignore_head, self.ignore_aggregates))
        # dprint('get_alpha_dict(%s)' % rule)
        # start = time.time()
        self.visit(rule)
        if self.tree_scope.get('invalidate'):
            dprint('WARN: Skipping invalid rule: %s' % rule, 1)
            self.variable_counter.clean() # Something went wrong or we found a rule that can't be processed, skip!
        self.variable_counter.post_process()
        end = time.time()
        # dprint('Dict Time: %f' % (end - start))
        # dict_total += end - start
        # dprint('Total Dict: %f' % dict_total, 0)
        dict_builder_cache[self] = self
        return self.variable_counter.variable_literal_dict

    def visit_children(self, x):
        for key in x.child_keys:
            if key == 'body':
                self.variable_counter.body_size = len(x[key])
            # Create a new/isolated scope so that children are not affected by siblings
            if isinstance(x[key], list):
                children = x[key]
                is_list = True
            else:
                children = [x[key]]
                is_list = False
            for index, child in enumerate(children): # Loop here to ensure a new scope is provided for immediate siblings
                child_tree_scope = dict(self.tree_scope)
                if key == 'head' or self.tree_scope.get('head'):
                    if self.ignore_head:
                        continue
                    child_tree_scope['head'] = True
                child_tree_scope['type'] = child.type if hasattr(child, 'type') else None
                child_tree_builder = AlphaDictionaryBuilder(self.base_transformer, self.variable_counter, child_tree_scope, self.tree_scope, self.ignore_head, self.ignore_aggregates)
                children[index] = child_tree_builder.visit(child)
            if is_list:
                x[key] = children
            else:
                x[key] = children[0]
        return x

    def visit_Aggregate(self, aggregate):
        self.tree_scope['aggregate'] = True
        return self.visit_children(aggregate)

    def visit_BodyAggregate(self, body_aggregate):
        self.tree_scope['aggregate'] = True
        return self.visit_children(body_aggregate)

    def visit_Literal(self, literal):
        self.tree_scope['literal'] = literal
        return self.visit_children(literal)
    
    def visit_ConditionalLiteral(self, cond_literal):
        if not self.tree_scope.get('head') and 'literal' not in self.tree_scope:
            self.invalidate() # Can't process conditional literals at the body root safely without more consideration
        return self.visit_children(cond_literal)

    def visit_Function(self, fun):
        if self.tree_scope.get('function'):
            self.tree_scope['function_arg'] = True # This is a function within function
        else:
            self.tree_scope['function'] = fun
        self.base_transformer.used_predicate_symbols.add(fun.name)
        return self.visit_children(fun)

    def visit_Variable(self, var):
        dprint('Checking: %s' % var, 3)
        self.variable_counter.all_vars.add(ASTWrapper(var))
        if self.tree_scope.get('aggregate'):
            self.variable_counter.aggregate_vars.add(ASTWrapper(var))
        else:
            self.variable_counter.non_aggregate_vars.add(ASTWrapper(var))
        if self.tree_scope.get('head') or (self.ignore_aggregates and self.tree_scope.get('aggregate')):
            dprint('Marking %s unsafe due to variable being inside head/aggregate' % var, 3)
            self.variable_counter.mark_var_unsafe(var)
        else:
            dprint('Adding %s to partition for %s' % (self.tree_scope.get('literal'), var), 3)
            self.variable_counter.add_literal_to_variable(var, self.tree_scope.get('literal'))
            if not self.tree_scope.get('function') or self.tree_scope.get('literal') and self.tree_scope.get('literal').sign != clingo.ast.Sign.NoSign:
                sibling_vars = clingo_ast_util.find_variables(self.tree_scope.get('literal'))
                self.variable_counter.create_link(sibling_vars)
        dprint('Checked: %s; Unsafe: %s' % (var, self.variable_counter.is_var_unsafe(var)), 3)
        return var
    
    def invalidate(self, scope=None):
        if scope is None:
            scope = self.tree_scope
        scope['invalidate'] = True
        parent_scope = scope.get('parent_scope')
        if parent_scope is not None:
            self.invalidate(parent_scope)

class VariableCounter:
    """This class is used to track how much a variable is used within a rule"""
    def __init__(self):
        self.body_size = -1
        self.clean()
    
    def clean(self):
        self.variable_literal_dict = {} # Maps a variable to a set of containing functions
        self.all_vars = set()
        self.aggregate_vars = set()
        self.non_aggregate_vars = set()
        self.unsafe_vars = set()
        self.unsafe_literals = set()
        self.variable_counts = {}
        self.safety_links = {}
    
    def post_process(self):
        dprint('DICT BEFORE LINK: %s' % self.variable_literal_dict, 3)
        dprint('Safety Links: %s' % self.safety_links, 3)
        new_literal_dict = {}
        for var in self.variable_literal_dict.keys():
            if var not in self.safety_links:
                new_literal_dict[var] = self.variable_literal_dict[var]
            else:
                new_literal_dict[var] = self.merge_safety_set(var)
        self.variable_literal_dict = new_literal_dict

        dprint('DICT AFTER LINK:  %s' % self.variable_literal_dict, 3)
        for unsafe_var in self.unsafe_vars:
            self.clean_var(unsafe_var.ast)
        for unsafe_literal in self.unsafe_literals:
            self.clean_literal(unsafe_literal.ast)
        dprint('DICT AFTER CLEAN: %s' % self.variable_literal_dict, 3)
    
    def merge_safety_set(self, var, done=None):
        if done is None:
            done = set()
        merged_set = set()
        for link_var in self.safety_links[var]:
            if link_var not in self.variable_literal_dict:
                raise Exception('Tried to merge links but no var key is in the alpha dict')
            if link_var in done:
                add_set = self.variable_literal_dict[link_var] # Base
            else:
                done.add(link_var)
                add_set = self.merge_safety_set(link_var, done)
            merged_set = merged_set.union(add_set)
        return merged_set

    def mark_var_unsafe(self, var):
        """Marks the variable as being unsafe for use in projection, removes variable from dict if already exists
        A variable should be unsafe if it matches any of the following: TODO Some of these conditions might be able to be worked around
        1.) The variable is in the head of the rule
        2.) The variable is used in an aggregate
        3.) The variable is used in an atom that is within another atom's arguments i.e. a(b(X))
        4.) The variable is used in all atoms of the body (therefore projection would not be of use)
        5.) The variable is used in greater than or equal to MAX_ATOM_PROJECITON_COUNT atoms in the body
        """
        dprint('Found unsafe var: %s' % var, 3)
        self.unsafe_vars.add(ASTWrapper(var))

    def mark_literal_unsafe(self, literal):
        dprint('Unsafe literal: %s' % literal, 3)
        self.unsafe_literals.add(ASTWrapper(literal))

    def clean_var(self, var):
        wrapped_var = ASTWrapper(var)
        if wrapped_var in self.variable_literal_dict:
            del self.variable_literal_dict[wrapped_var]
    
    def clean_literal(self, literal):
        wrapped_literal = ASTWrapper(literal)
        for wrapped_var, wrapped_literal_set in dict(self.variable_literal_dict).items():
            size_before = len(wrapped_literal_set)
            wrapped_literal_set = set([o_wrapped_literal for o_wrapped_literal in wrapped_literal_set if o_wrapped_literal != wrapped_literal])
            size_after = len(wrapped_literal_set)
            if size_after < size_before:
                # Function was removed, this variable is also unsafe
                dprint('Marking %s unsafe due to literal (%s) being unsafe' % (wrapped_var.ast, literal), 3)
                self.mark_var_unsafe(wrapped_var.ast)
                if len(wrapped_literal_set) > 0:
                    self.variable_literal_dict[wrapped_var] = wrapped_literal_set
    
    def is_var_unsafe(self, var):
        """Returns whether or not var is unsafe"""
        return ASTWrapper(var) in self.unsafe_vars

    def is_literal_unsafe(self, literal):
        """Returns whether or not literal is unsafe"""
        return ASTWrapper(literal) in self.unsafe_literals
    
    def add_literal_to_variable(self, var, literal):
        """Adds a function to the set of containing functions for a variable
        If it detects that this variable is used across the entire body
        (by comparing the # of functions in the partition to the body size)
        or that the variable is shared across MAX_ATOM_PROJECTION_COUNT literals,
        it will mark the variable as unsafe instead
        """
        wrapped_var = ASTWrapper(var)
        if wrapped_var not in self.variable_literal_dict:
            self.variable_literal_dict[wrapped_var] = set()
        if len(self.variable_literal_dict[wrapped_var]) >= pr_settings.args.max_atoms:
            dprint('Marking %s unsafe due to exceeding MAX_ATOM_PROJECTION_COUNT' % var, 3)
            self.mark_var_unsafe(var)
        
        self.variable_literal_dict[wrapped_var].add(ASTWrapper(literal))
        return True

    def create_link(self, var_set):
        dprint('Link created for vars: %s' % var_set, 3)
        for var in var_set:
            if var in self.safety_links:
                self.safety_links[var] = self.safety_links[var].union(var_set) # Var already has links, union all links
            else:
                self.safety_links[var] = var_set # First link for var

    def get_aggregate_ignore_vars(self):
        """Returns a set of variables corresponding to those that are both outside of aggregates and inside aggregates"""
        # TODO This is an issue if there are multiple aggregates
        return self.aggregate_vars.intersection(self.non_aggregate_vars)
