import argparse, sys, random

def init():
    global args, answer_queue, oracle_done, heuristic_random, rule_random
    VERSION = '1.0.0'
    output_list = ['ground', 'solve', 'rewrite']
    heuristic_list = ['variable', 'ascending', 'descending']

    parser = argparse.ArgumentParser(description='Rewrites ASP logic program')
    parser.add_argument('encoding', nargs='+', default=[], help='gringo source files')
    parser.add_argument('--instance', action='append', help='adds an additional file to include in grounding/solving (will not be rewritten)')
    parser.add_argument('--output', type=str, default='rewrite', help='the output mode of the program [%s]' % '|'.join(output_list))
    parser.add_argument('--debug', type=int, default=0, help='enables pausing after projection & debug trace messages, value controls the level, 0 disables')
    parser.add_argument('--max_atoms', type=int, default=-1, help='the maximum number of atoms a variable can be shared with in a rule while still performing projection, -1 for no max atoms')
    parser.add_argument('--restrict', type=str, default='alpha', help='the restrict mode of the program')
    parser.add_argument('--no_normalize', action='store_true', help='normalizes projected rules (NOTE: breaks caching for differently named variables)')
    parser.add_argument('--limit', type=int, default=float('inf'), help='limits the number of projections performed, stopping projection after encountering the number specified')
    parser.add_argument('--debug_after', type=int, default=float('inf'), help='enables debug mode after specified number of projections')
    parser.add_argument('--random', type=int, default=0, help='provides a seed for randomness on equal heuristics, 0 to disable and always choose the first encountered (default)')
    parser.add_argument('--random_rules', type=int, default=0, help='provides a seed for random choice selection on whether or not to perform a rewrite, 0 to disable (default)')
    parser.add_argument('--oracle', action='store_true', help='predicts projections based on grounding size (will only improve solve time; grounding will occur many times)')
    parser.add_argument('--no_rewrite', action='store_true', help='disables all rewriting and just parses the given program')
    # parser.add_argument('--descending', action='store_true', help='makes the alpha set selection choose LARGEST sets first')
    parser.add_argument('--heuristic', type=str, default='ascending', help='the variable selection heuristic to use [%s]' % '|'.join(heuristic_list))
    parser.add_argument('--aggregates', action='store_true', help='enable projection on aggregates')
    parser.add_argument('--old_mode', action='store_true', help='sets certain changes back to the state they were in during the paper (currently just beta literal removal from replacement rule)')
    parser.add_argument('--conservative', action='store_true', help='enables conservative projection mode')
    parser.add_argument('--predictor', action='store_true', help='enables the use of Predictor for variable heuristics')
    parser.add_argument('--key', action='append', help='specifies the predicate keys to use with predictor (predictor must be enabled; format: p/3[1,3])')
    parser.add_argument('--version', action='store_true', help='prints the version of Projector')

    args = parser.parse_args()
    if args.version:
        print('Projector version %s' % VERSION)
        sys.exit(0)
    if args.output not in output_list or args.heuristic not in heuristic_list:
        parser.print_help()
        sys.exit(0)
    if args.max_atoms < 0:
        args.max_atoms = float('inf')
    
    answer_queue = [] # Used for oracle
    oracle_done = False
    heuristic_random = random.Random()
    if args.random:
        heuristic_random.seed(args.random)
    rule_random = random.Random()
    if args.random_rules:
        rule_random.seed(args.random_rules)
